@@ -1,5 +1,5 @@
 #
-# Copyright (c), 2018-2019, SISSA (International School for Advanced Studies).
+# Copyright (c), 2018-2020, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -17,16 +17,18 @@ import time
 import re
 import locale
 import unicodedata
+from copy import copy
 from urllib.parse import urlparse, urljoin, quote as urllib_quote
-from urllib.error import URLError
 import xml.etree.ElementTree as ElementTree
 
-from .datatypes import QNAME_PATTERN, DateTime10, Date10, Time, Timezone, Duration, DayTimeDuration
-from .namespaces import prefixed_to_qname, get_namespace, XML_ID
+from .exceptions import ElementPathTypeError, xpath_error
+from .datatypes import QNAME_PATTERN, DateTime10, Date10, Time, Timezone, \
+    Duration, DayTimeDuration, is_id, is_idrefs
+from .namespaces import get_namespace, XML_ID
 from .xpath_context import XPathContext, XPathSchemaContext
-from .xpath_nodes import AttributeNode, is_document_node, is_xpath_node, is_element_node, \
-    is_attribute_node, node_name, node_is_id, node_is_idrefs, node_nilled, node_base_uri, \
-    node_document_uri, node_kind, etree_deep_equal
+from .xpath_nodes import AttributeNode, is_document_node, is_xpath_node, \
+    is_element_node, is_attribute_node, node_name, node_nilled, \
+    node_base_uri, node_document_uri, node_kind, etree_deep_equal
 from .xpath2_parser import XPath2Parser
 
 method = XPath2Parser.method
@@ -37,10 +39,7 @@ WRONG_REPLACEMENT_PATTERN = re.compile(r'(?<!\\)\$([^\d]|$)|((?<=[^\\])|^)\\([^$
 
 ###
 # Sequence types (allowed only for type checking in treat-as/instance-of statements)
-@method(function('empty-sequence', nargs=0, label='sequence type'))
-def evaluate(self, context=None):
-    if context is not None:
-        return isinstance(context.item, list) and not context.item
+function('empty-sequence', nargs=0, label='sequence type')
 
 
 @method(function('item', nargs=0, label='sequence type'))
@@ -53,70 +52,6 @@ def evaluate(self, context=None):
         return context.item
 
 
-###
-# Kind tests (sequence types that can appear also in XPath expressions)
-@method(function('document-node', nargs=(0, 1), label='kind test'))
-def evaluate(self, context=None):
-    if context is not None:
-        if context.item is None and is_document_node(context.root):
-            if not self:
-                return context.root
-            elif is_element_node(context.root.getroot(), self[0].evaluate(context)):
-                return context.root
-
-
-@method(function('element', nargs=(0, 2), label='kind test'))
-def evaluate(self, context=None):
-    if context is not None:
-        if not self:
-            if is_element_node(context.item):
-                return context.item
-        elif is_element_node(context.item, self[0].evaluate(context)):
-            if len(self) == 1:
-                return context.item
-            elif self.xsd_type is not None:
-                type_annotation = self[1].evaluate(context)
-                if self.xsd_type.is_matching(type_annotation, self.parser.default_namespace):
-                    return context.item
-
-
-@method(function('schema-attribute', nargs=1, label='kind test'))
-def evaluate(self, context=None):
-    attribute_name = self[0].source
-    qname = prefixed_to_qname(attribute_name, self.parser.namespaces)
-    if self.parser.schema.get_attribute(qname) is None:
-        self.missing_name("attribute %r not found in schema" % attribute_name)
-
-    if context is not None:
-        if is_attribute_node(context.item, qname):
-            return context.item
-
-
-@method(function('schema-element', nargs=1, label='kind test'))
-def evaluate(self, context=None):
-    element_name = self[0].source
-    qname = prefixed_to_qname(element_name, self.parser.namespaces)
-    if self.parser.schema.get_element(qname) is None \
-            and self.parser.schema.get_substitution_group(qname) is None:
-        self.missing_name("element %r not found in schema" % element_name)
-
-    if context is not None:
-        if is_element_node(context.item) and context.item.tag == qname:
-            return context.item
-
-
-@method('document-node')
-@method('element')
-@method('schema-attribute')
-@method('schema-element')
-def select(self, context=None):
-    if context is not None:
-        for _ in context.iter_children_or_self():
-            item = self.evaluate(context)
-            if item is not None:
-                yield item
-
-
 ###
 # Function for QNames
 @method(function('prefix-from-QName', nargs=1))
@@ -166,7 +101,7 @@ def evaluate(self, context=None):
 @method(function('namespace-uri-for-prefix', nargs=2))
 def evaluate(self, context=None):
     if context is not None:
-        prefix = self.get_argument(context.copy())
+        prefix = self.get_argument(context=copy(context))
         if prefix is None:
             prefix = ''
         if not isinstance(prefix, str):
@@ -182,7 +117,8 @@ def evaluate(self, context=None):
                     if not prefix or uri:
                         return uri
                     else:
-                        raise self.error('XPST0081', 'Prefix %r is associated to no namespace' % prefix)
+                        msg = 'Prefix %r is associated to no namespace'
+                        raise self.error('XPST0081', msg % prefix)
         return []
 
 
@@ -215,7 +151,7 @@ def evaluate(self, context=None):
         if not is_element_node(elem):
             raise self.error('FORG0006', '2nd argument %r is not a node' % elem)
 
-        qname = self.get_argument(context.copy())
+        qname = self.get_argument(context=copy(context))
         if qname is None:
             return []
         if not isinstance(qname, str):
@@ -226,10 +162,12 @@ def evaluate(self, context=None):
         prefix = match.groupdict()['prefix'] or ''
 
         ns_uris = {get_namespace(e.tag) for e in elem.iter()}
-        for p, uri in self.parser.namespaces.items():
-            if uri in ns_uris:
-                if not prefix or uri:
-                    return '{%s}%s' % (uri, match.groupdict()['local']) if uri else match.groupdict()['local']
+        for pfx, uri in self.parser.namespaces.items():
+            if pfx == prefix and uri in ns_uris:
+                if uri:
+                    return '{%s}%s' % (uri, match.groupdict()['local'])
+                elif not prefix:
+                    return match.groupdict()['local']
                 else:
                     raise self.error('XPST0081', 'Prefix %r is associated to no namespace' % prefix)
 
@@ -255,26 +193,27 @@ def select(self, context=None):
     for item in self[0].select(context):
         value = self.data_value(item)
         if value is None:
-            raise self.error('FOTY0012', "argument node does not have a typed value: %r" % item)
+            raise self.error('FOTY0012', "argument node does not have a typed value")
         else:
             yield value
 
 
 @method(function('base-uri', nargs=(0, 1)))
 def evaluate(self, context=None):
-    item = self.get_argument(context)
+    item = self.get_argument(context, default_to_context=True)
     if item is None:
         self.missing_context("context item is undefined")
     elif not is_xpath_node(item):
         self.wrong_context_type("context item is not a node")
     else:
-        return node_base_uri
+        return node_base_uri(item)
 
 
 @method(function('document-uri', nargs=1))
 def evaluate(self, context=None):
-    arg = self.get_argument(context)
-    return [] if arg is None else node_document_uri(arg)
+    if context is not None:
+        arg = self.get_argument(context)
+        return [] if arg is None else node_document_uri(arg)
 
 
 ###
@@ -286,16 +225,11 @@ def evaluate(self, context=None):
         return []
     elif isinstance(item, float) and (math.isnan(item) or math.isinf(item)):
         return item
+    elif not isinstance(item, (float, int, decimal.Decimal)):
+        self.wrong_type("Invalid argument type {!r}".format(type(item)))
 
-    try:
-        precision = 0 if len(self) < 2 else self[1].evaluate(context)
-        value = round(decimal.Decimal(item), precision)
-    except TypeError as err:
-        self.wrong_type(str(err))
-    except decimal.DecimalException as err:
-        self.wrong_value(str(err))
-    else:
-        return float(value)
+    precision = 0 if len(self) < 2 else self[1].evaluate(context)
+    return float(round(decimal.Decimal(item), precision))
 
 
 @method(function('abs', nargs=1))
@@ -305,11 +239,16 @@ def evaluate(self, context=None):
         return []
     elif isinstance(item, float) and math.isnan(item):
         return item
-
-    try:
-        return abs(self.string_value(item) if is_xpath_node(item) else item)
-    except TypeError as err:
-        self.wrong_type(str(err))
+    elif is_xpath_node(item):
+        value = self.string_value(item)
+        try:
+            return abs(decimal.Decimal(value))
+        except decimal.DecimalException:
+            self.wrong_value("Invalid string value {!r} for {!r}".format(value, item))
+    elif not isinstance(item, (float, int, decimal.Decimal)):
+        self.wrong_type("Invalid argument type {!r}".format(type(item)))
+    else:
+        return abs(item)
 
 
 ###
@@ -337,13 +276,15 @@ def evaluate(self, context=None):
 @method(function('max', nargs=(1, 2)))
 @method(function('min', nargs=(1, 2)))
 def evaluate(self, context=None):
-    arg = self[0].select(context)
     func = max if self.symbol == 'max' else min
     try:
+        values = [x for x in self[0].select(context)]
+        if any(isinstance(x, float) and math.isnan(x) for x in values):
+            return float('nan')
         if len(self) > 1:
             with self.use_locale(collation=self.get_argument(context, 1)):
-                return func(arg)
-        return func(arg)
+                return func(values)
+        return func(values)
     except TypeError as err:
         self.wrong_type(str(err))
     except ValueError:
@@ -519,12 +460,8 @@ def evaluate(self, context=None):
             elif not etree_deep_equal(value1, value2):
                 return False
 
-    if context is None:
-        seq1 = iter(self[0].select())
-        seq2 = iter(self[1].select())
-    else:
-        seq1 = iter(self[0].select(context.copy()))
-        seq2 = iter(self[1].select(context.copy()))
+    seq1 = iter(self[0].select(copy(context)))
+    seq2 = iter(self[1].select(copy(context)))
 
     if len(self) > 2:
         with self.use_locale(collation=self.get_argument(context, 2)):
@@ -550,7 +487,8 @@ def evaluate(self, context=None):
     try:
         return re.search(pattern, input_string, flags=flags) is not None
     except re.error:
-        raise self.error('FORX0002', "Invalid regular expression %r" % pattern)  # TODO: full XML regex syntax
+        # TODO: full XML regex syntax
+        raise self.error('FORX0002', "Invalid regular expression %r" % pattern)
 
 
 @method(function('replace', nargs=(3, 4)))
@@ -569,10 +507,12 @@ def evaluate(self, context=None):
     try:
         pattern = re.compile(pattern, flags=flags)
     except re.error:
-        raise self.error('FORX0002', "Invalid regular expression %r" % pattern)  # TODO: full XML regex syntax
+        # TODO: full XML regex syntax
+        raise self.error('FORX0002', "Invalid regular expression %r" % pattern)
     else:
         if pattern.search(''):
-            raise self.error('FORX0003', "Regular expression %r matches zero-length string" % pattern.pattern)
+            msg = "Regular expression %r matches zero-length string"
+            raise self.error('FORX0003', msg % pattern.pattern)
         elif WRONG_REPLACEMENT_PATTERN.search(replacement):
             raise self.error('FORX0004', "Invalid replacement string %r" % replacement)
         else:
@@ -601,7 +541,8 @@ def select(self, context=None):
         raise self.error('FORX0002', "Invalid regular expression %r" % pattern)
     else:
         if pattern.search(''):
-            raise self.error('FORX0003', "Regular expression %r matches zero-length string" % pattern.pattern)
+            msg = "Regular expression %r matches zero-length string"
+            raise self.error('FORX0003', msg % pattern.pattern)
 
     if input_string:
         for value in pattern.split(input_string):
@@ -616,7 +557,10 @@ def evaluate(self, context=None):
     relative = self.get_argument(context, cls=str)
     if len(self) == 2:
         base_uri = self.get_argument(context, index=1, required=True, cls=str)
-        base_uri = urlparse(base_uri).geturl()
+        url_parts = urlparse(base_uri)
+        if not url_parts.path.startswith('/'):
+            raise self.error('FORG0002', '2nd argument is not an absolute URI')
+        base_uri = url_parts.geturl()
     elif self.parser.base_uri is None:
         raise self.error('FONS0005')
     else:
@@ -653,13 +597,12 @@ def evaluate(self, context=None):
         return []
 
     if len(self) < 3:
-        locale.setlocale(locale.LC_ALL, '')
         value = locale.strcoll(comp1, comp2)
     else:
         with self.use_locale(collation=self.get_argument(context, 2)):
             value = locale.strcoll(comp1, comp2)
 
-    return 1 if value > 0 else -1 if value < 0 else 0
+    return 0 if not value else 1 if value > 0 else -1
 
 
 @method(function('codepoint-equal', nargs=2))
@@ -674,14 +617,16 @@ def evaluate(self, context=None):
         return all(ord(c1) == ord(c2) for c1, c2 in zip(comp1, comp2))
 
 
-@method(function('string-join', nargs=2))
+@method(function('string-join', nargs=(1, 2)))
 def evaluate(self, context=None):
     items = [self.string_value(s) if is_element_node(s) or is_attribute_node(s) else s
              for s in self[0].select(context)]
     try:
-        return self.get_argument(context, 1, cls=str).join(items)
-    except AttributeError as err:
-        self.wrong_type("the separator must be a string: %s" % err)
+        return self.get_argument(context, 1, default='', cls=str).join(items)
+    except AttributeError:
+        return ''.join(items)
+    except ElementPathTypeError:
+        raise
     except TypeError as err:
         self.wrong_type("the values must be strings: %s" % err)
 
@@ -700,10 +645,8 @@ def evaluate(self, context=None):
 
     if normalization_form == 'FULLY-NORMALIZED':
         raise NotImplementedError("%r normalization form not supported" % normalization_form)
-    if arg is None:
+    if not arg:
         return ''
-    elif not isinstance(arg, str):
-        arg = arg.decode('utf-8')
 
     try:
         return unicodedata.normalize(normalization_form, arg)
@@ -713,47 +656,32 @@ def evaluate(self, context=None):
 
 @method(function('upper-case', nargs=1))
 def evaluate(self, context=None):
-    arg = self.get_argument(context, cls=str)
-    try:
-        return '' if arg is None else arg.upper()
-    except AttributeError:
-        self.wrong_type("the argument must be a string: %r" % arg)
+    return self.get_argument(context, default='', cls=str).upper()
 
 
 @method(function('lower-case', nargs=1))
 def evaluate(self, context=None):
-    arg = self.get_argument(context, cls=str)
-    try:
-        return '' if arg is None else arg.lower()
-    except AttributeError:
-        self.wrong_type("the argument must be a string: %r" % arg)
+    return self.get_argument(context, default='', cls=str).lower()
 
 
 @method(function('encode-for-uri', nargs=1))
 def evaluate(self, context=None):
     uri_part = self.get_argument(context, cls=str)
-    try:
-        return '' if uri_part is None else urllib_quote(uri_part, safe='~')
-    except TypeError:
-        self.wrong_type("the argument must be a string: %r" % uri_part)
+    return '' if uri_part is None else urllib_quote(uri_part, safe='~')
 
 
 @method(function('iri-to-uri', nargs=1))
 def evaluate(self, context=None):
     iri = self.get_argument(context, cls=str)
-    try:
-        return '' if iri is None else urllib_quote(iri, safe='-_.!~*\'()#;/?:@&=+$,[]%')
-    except TypeError:
-        self.wrong_type("the argument must be a string: %r" % iri)
+    return '' if iri is None else urllib_quote(iri, safe='-_.!~*\'()#;/?:@&=+$,[]%')
 
 
 @method(function('escape-html-uri', nargs=1))
 def evaluate(self, context=None):
     uri = self.get_argument(context, cls=str)
-    try:
-        return '' if uri is None else urllib_quote(uri, safe=''.join(chr(cp) for cp in range(32, 127)))
-    except TypeError:
-        self.wrong_type("the argument must be a string: %r" % uri)
+    if uri is None:
+        return ''
+    return urllib_quote(uri, safe=''.join(chr(cp) for cp in range(32, 127)))
 
 
 @method(function('starts-with', nargs=(2, 3)))
@@ -951,7 +879,8 @@ def evaluate(self, context=None):
 @method(function('current-dateTime', nargs=0))
 def evaluate(self, context=None):
     dt = datetime.datetime.now() if context is None else context.current_dt
-    return DateTime10(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.microsecond, dt.tzinfo)
+    return DateTime10(dt.year, dt.month, dt.day, dt.hour, dt.minute,
+                      dt.second, dt.microsecond, dt.tzinfo)
 
 
 @method(function('current-date', nargs=0))
@@ -1009,9 +938,11 @@ XPath2Parser.unregister('id')
 @method(function('id', nargs=(1, 2)))
 def select(self, context=None):
     # TODO: PSVI bindings with also xsi:type evaluation
-    idrefs = list(self[0].select(None if context is None else context.copy()))
+    idrefs = [x for x in self[0].select(context=copy(context))]
     node = self.get_argument(context, index=1, default_to_context=True)
-    if context is None or node is not context.item:
+    if context is None:
+        return
+    elif node is not context.item:
         if not is_document_node(node):
             raise self.error('FODC0001', 'cannot retrieve document root')
         root = node
@@ -1023,7 +954,7 @@ def select(self, context=None):
         root = context.root
 
     for elem in root.iter():
-        if node_is_id(elem) and any(v == elem.text for x in idrefs for v in x.split()):
+        if is_id(elem.text) and any(v == elem.text for x in idrefs for v in x.split()):
             yield elem
             continue
         for attr in map(lambda x: AttributeNode(*x), elem.attrib.items()):
@@ -1035,9 +966,11 @@ def select(self, context=None):
 @method(function('idref', nargs=(1, 2)))
 def select(self, context=None):
     # TODO: PSVI bindings with also xsi:type evaluation
-    ids = list(self[0].select(None if context is None else context.copy()))
+    ids = [x for x in self[0].select(context=copy(context))]
     node = self.get_argument(context, index=1, default_to_context=True)
-    if context is None or node is not context.item:
+    if context is None:
+        return
+    elif node is not context.item:
         if not is_document_node(node):
             raise self.error('FODC0001', 'cannot retrieve document root')
         root = node
@@ -1049,7 +982,7 @@ def select(self, context=None):
         root = context.root
 
     for elem in root.iter():
-        if node_is_idrefs(elem) and any(v in elem.text.split() for x in ids for v in x.split()):
+        if is_idrefs(elem.text) and any(v in elem.text.split() for x in ids for v in x.split()):
             yield elem
             continue
         for attr in map(lambda x: AttributeNode(*x), elem.attrib.items()):
@@ -1064,11 +997,7 @@ def evaluate(self, context=None):
     uri = self.get_argument(context)
     if uri is None:
         return None if self.symbol == 'doc' else False
-
-    try:
-        uri = self.get_absolute_uri(uri)
-    except URLError:
-        raise self.error('FODC0005')
+    uri = self.get_absolute_uri(uri)
 
     if context is not None and not isinstance(context, XPathSchemaContext):
         try:
@@ -1093,11 +1022,7 @@ def evaluate(self, context=None):
             raise self.error('FODC0002', 'no default collection has been defined')
         return context.default_collection
 
-    try:
-        uri = self.get_absolute_uri(uri)
-    except URLError:
-        raise self.error('FODC0004')
-
+    uri = self.get_absolute_uri(uri)
     try:
         return self.parser.collections[uri]
     except KeyError:
@@ -1109,10 +1034,14 @@ def evaluate(self, context=None):
 @method(function('error', nargs=(0, 3)))
 def evaluate(self, context=None):
     if not self:
-        raise self.error('FOER0000')
+        raise xpath_error('FOER0000')
     elif len(self) == 1:
-        item = self.get_argument(context)
-        raise self.error(item or 'FOER0000')
+        error = self.get_argument(context, cls=str)
+        raise xpath_error(error or 'FOER0000')
+    else:
+        error = self.get_argument(context, cls=str)
+        description = self.get_argument(context, index=1, cls=str)
+        raise xpath_error(error or 'FOER0000', message=description)
 
 
 # XPath 2.0 definitions continue into module xpath2_constructors
