@@ -27,10 +27,11 @@ from .exceptions import xpath_error
 from .namespaces import XQT_ERRORS_NAMESPACE, XSD_NAMESPACE, XPATH_FUNCTIONS_NAMESPACE, \
     XSD_ANY_TYPE, XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE
 from .xpath_nodes import AttributeNode, TextNode, TypedAttribute, TypedElement, \
-    is_etree_element, is_attribute_node, elem_iter_strings, is_text_node, \
+    is_etree_element, is_attribute_node, etree_iter_strings, is_text_node, \
     is_namespace_node, is_comment_node, is_processing_instruction_node, \
     is_element_node, is_document_node, is_xpath_node, is_schema_node
-from .datatypes import UntypedAtomic, Timezone, DayTimeDuration, XSD_BUILTIN_TYPES
+from .datatypes import UntypedAtomic, Timezone, DateTime10, Date10, \
+    DayTimeDuration, XSD_BUILTIN_TYPES
 from .schema_proxy import AbstractSchemaProxy
 from .tdop_parser import Token, MultiLabel
 from .xpath_context import XPathSchemaContext
@@ -59,7 +60,7 @@ def ordinal(n):
 class XPathToken(Token):
     """Base class for XPath tokens."""
     comment = None    # for XPath 2.0+ comments
-    xsd_type = None   # fox XPath 2.0+ schema types labeling
+    xsd_types = None   # fox XPath 2.0+ schema types labeling
 
     def evaluate(self, context=None):
         """
@@ -240,12 +241,15 @@ class XPathToken(Token):
                     value = str(value)
                 if isinstance(context, XPathSchemaContext):
                     return value
-                if self.xsd_type is not None and isinstance(value, str):
-                    if self.xsd_type.name in XSD_SPECIAL_TYPES:
+                if self.xsd_types and isinstance(value, str):
+                    xsd_type = self.get_xsd_type(context.item)
+                    if xsd_type is None:
+                        pass
+                    elif xsd_type.name in XSD_SPECIAL_TYPES:
                         value = UntypedAtomic(value)
                     else:
                         try:
-                            value = self.xsd_type.decode(value)
+                            value = xsd_type.decode(value)
                         except (TypeError, ValueError):
                             msg = "Type {!r} is not appropriate for the context"
                             self.wrong_context_type(msg.format(type(value)))
@@ -416,7 +420,7 @@ class XPathToken(Token):
         if len(self) == 1:
             item = self.get_argument(context, cls=cls)
             if item is None:
-                return []
+                return
             timezone = getattr(context, 'timezone', None)
         else:
             item = self.get_argument(context=None, cls=cls)  # don't use implicit timezone
@@ -425,100 +429,17 @@ class XPathToken(Token):
                 timezone = Timezone.fromduration(timezone)
 
         if item.tzinfo is not None and timezone is not None:
-            item += timezone.offset - item.tzinfo.offset
-            item.tzinfo = timezone
-        elif item.tzinfo is None:
-            if timezone is not None:
-                item.tzinfo = timezone
-        elif timezone is None:
-            item.tzinfo = None
+            if isinstance(item, DateTime10):
+                item += timezone.offset
+            elif not isinstance(item, Date10):
+                item += timezone.offset - item.tzinfo.offset
+            elif timezone.offset < item.tzinfo.offset:
+                item -= timezone.offset - item.tzinfo.offset
+                item -= DayTimeDuration.fromstring('P1D')
+
+        item.tzinfo = timezone
         return item
 
-    def match_xsd_type(self, schema_item, name):
-        """
-        Match a token with a schema type, checking the matching between the provided schema
-        item and name. If there is a match and the token is already related with another
-        schema type an exception is raised.
-
-        :param schema_item: an XPath item related with a schema instance.
-        :param name: a QName in extended format for matching the item.
-        :returns: the matched XSD type or `None` if there isn't a match.
-        """
-        if isinstance(schema_item, AttributeNode):
-            if not schema_item[1].is_matching(name):
-                return
-
-            try:
-                xsd_type = schema_item[1].type
-            except AttributeError:
-                try:
-                    xsd_type = self.parser.schema.get_attribute(name).type
-                except AttributeError:
-                    return
-
-        elif is_etree_element(schema_item):
-            if hasattr(schema_item, 'is_matching'):
-                if not schema_item.is_matching(name, self.parser.default_namespace):
-                    return
-            elif schema_item.tag != name:
-                return
-
-            try:
-                xsd_type = schema_item.type
-            except AttributeError:
-                try:
-                    xsd_type = self.parser.schema.get_element(name).type
-                except AttributeError:
-                    return
-        else:
-            return
-
-        if self.xsd_type is None:
-            self.xsd_type = xsd_type
-        elif self.xsd_type is not xsd_type:
-            self.wrong_context_type("Multiple XSD type matching during static analysis")
-        return xsd_type
-
-    def get_typed_node(self, context, item):
-        """
-        Returns a typed node if the token is bound to an XSD type.
-
-        :param context: the XPath dynamic context.
-        :param item: an untyped XPath attribute ot element.
-        """
-        if isinstance(self.xsd_type, (type(None), AbstractSchemaProxy)):
-            return item
-
-        if isinstance(context, XPathSchemaContext):
-            primitive_type = self.parser.schema.get_primitive_type(self.xsd_type)
-            try:
-                value = XSD_BUILTIN_TYPES[primitive_type.local_name or 'anyType'].value
-            except KeyError:
-                value = XSD_BUILTIN_TYPES['anyType'].value
-
-            if isinstance(item, AttributeNode):
-                return TypedAttribute(item, value)
-            else:
-                return TypedElement(item, value)
-        else:
-            try:
-                if isinstance(item, AttributeNode):
-                    if self.xsd_type.name in XSD_SPECIAL_TYPES:
-                        return TypedAttribute(item, UntypedAtomic(item[1]))
-                    else:
-                        return TypedAttribute(item, self.xsd_type.decode(item[1]))
-
-                elif self.xsd_type.is_simple() or self.xsd_type.has_simple_content():
-                    if self.xsd_type.name in XSD_SPECIAL_TYPES:
-                        return TypedElement(item, UntypedAtomic(item.text))
-                    else:
-                        return TypedElement(item, self.xsd_type.decode(item.text))
-                else:
-                    return item
-            except (TypeError, ValueError):
-                msg = "Type {!r} does not match sequence type of {!r}"
-                self.wrong_sequence_type(msg.format(self.xsd_type, item))
-
     @contextlib.contextmanager
     def use_locale(self, collation):
         """A context manager for use a locale setting for string comparison in a code block."""
@@ -535,23 +456,159 @@ class XPathToken(Token):
         finally:
             locale.setlocale(locale.LC_COLLATE, loc)
 
+    ###
+    # XSD type association helper methods
+    def add_xsd_type(self, name, xsd_type):
+        """
+        Adds an XSD type association to token.
+
+        :param name: the name to match with the XSD type.
+        :param xsd_type: the XSD type to add.
+        """
+        if self.xsd_types is None:
+            self.xsd_types = {name: xsd_type}
+        else:
+            obj = self.xsd_types.get(name)
+            if obj is None:
+                self.xsd_types[name] = xsd_type
+            elif not isinstance(obj, list):
+                if obj is not xsd_type:
+                    self.xsd_types[name] = [obj, xsd_type]
+            elif xsd_type not in obj:
+                obj.append(xsd_type)
+
+        return xsd_type
+
+    def match_xsd_type(self, schema_item, name):
+        """
+        Match a token with a schema type, checking the matching between the provided schema
+        item and name. If there is a match and the token is already related with another
+        schema type an exception is raised.
+
+        :param schema_item: an XPath item related with a schema instance.
+        :param name: a QName in extended format for matching the item.
+        :returns: the matched XSD type or `None` if there isn't a match.
+        """
+        try:
+            if isinstance(schema_item, AttributeNode):
+                if not schema_item[1].is_matching(name):
+                    return
+                xsd_type = schema_item[1].type
+            elif not schema_item.is_matching(name, self.parser.default_namespace):
+                return
+            else:
+                xsd_type = schema_item.type
+        except AttributeError:
+            return
+
+        self.add_xsd_type(name, xsd_type)
+
+        try:
+            value = XSD_BUILTIN_TYPES[xsd_type.local_name].value
+        except KeyError:
+            primitive_type = self.parser.schema.get_primitive_type(xsd_type)
+            try:
+                value = XSD_BUILTIN_TYPES[primitive_type.local_name or 'anyType'].value
+            except KeyError:
+                value = XSD_BUILTIN_TYPES['anyType'].value
+
+        if isinstance(schema_item, AttributeNode):
+            return TypedAttribute(schema_item, value)
+        return TypedElement(schema_item, value)
+
+    def get_xsd_type(self, item):
+        """
+        Returns the XSD type associated with an item. Match by item's name
+        and XSD validity. Returns `None` if no XSD type is matching.
+
+        :param item: a string or an AttributeNode or an element.
+        """
+        if not self.xsd_types or isinstance(self.xsd_types, AbstractSchemaProxy):
+            return
+        elif isinstance(item, str):
+            xsd_type = self.xsd_types.get(item)
+        elif isinstance(item, AttributeNode):
+            xsd_type = self.xsd_types.get(item[0])
+        else:
+            xsd_type = self.xsd_types.get(item.tag)
+
+        if not xsd_type:
+            return
+        elif not isinstance(xsd_type, list):
+            return xsd_type
+        elif isinstance(item, AttributeNode):
+            for x in xsd_type:
+                if x.is_valid(item[1]):
+                    return x
+        elif not isinstance(item, str):
+            for x in xsd_type:
+                if x.is_simple():
+                    if x.is_valid(item.text):
+                        return x
+                elif x.is_valid(item):
+                    return x
+
+        return xsd_type[0]
+
+    def get_typed_node(self, item):
+        """
+        Returns a typed node if the item is matching an XSD type.
+
+        Ref:
+          https://www.w3.org/TR/xpath20/#id-processing-model
+          https://www.w3.org/TR/xpath20/#id-static-analysis
+          https://www.w3.org/TR/xquery-semantics/
+
+        :param item: an untyped attribute ot element.
+        :return: a TypedAttribute or a TypedElement, or the argument \
+        if it's not matching any associated XSD type.
+        """
+        xsd_type = self.get_xsd_type(item)
+        if not xsd_type:
+            return item
+
+        try:
+            if isinstance(item, AttributeNode):
+                if xsd_type.name in XSD_SPECIAL_TYPES:
+                    return TypedAttribute(item, UntypedAtomic(item[1]))
+                else:
+                    return TypedAttribute(item, xsd_type.decode(item[1]))
+
+            elif xsd_type.is_simple() or xsd_type.has_simple_content():
+                if xsd_type.name in XSD_SPECIAL_TYPES:
+                    return TypedElement(item, UntypedAtomic(item.text))
+                else:
+                    return TypedElement(item, xsd_type.decode(item.text))
+            else:
+                return item
+
+        except (TypeError, ValueError):
+            msg = "Type {!r} does not match sequence type of {!r}"
+            self.wrong_sequence_type(msg.format(xsd_type, item))
+
     ###
     # XPath data accessors base functions
     def data_value(self, obj):
         """
-        The typed value, as computed by fn:data() on each item. Returns an instance of
-        UntypedAtomic.
+        The typed value, as computed by fn:data() on each item.
+        Returns an instance of UntypedAtomic for untyped data.
         """
-        if is_attribute_node(obj) or isinstance(obj, TypedElement):
-            obj = obj[1]
-
         if obj is None:
             return
-        elif not is_xpath_node(obj):
-            return obj
         elif hasattr(obj, 'type'):
             return self.schema_node_value(obj)  # Schema context
-        return UntypedAtomic(self.string_value(obj))
+        elif isinstance(obj, (TypedElement, TypedAttribute)):
+            return obj[-1]
+        elif isinstance(obj, tuple):
+            return UntypedAtomic(obj[-1])
+        elif is_etree_element(obj):
+            value = ''.join(etree_iter_strings(obj))
+            return UntypedAtomic(value)
+        elif is_document_node(obj):
+            value = ''.join(etree_iter_strings(obj.getroot()))
+            return UntypedAtomic(value)
+        else:
+            return obj
 
     def boolean_value(self, obj):
         """
@@ -560,7 +617,13 @@ class XPathToken(Token):
         if isinstance(obj, list):
             if not obj:
                 return False
-            elif isinstance(obj[0], tuple) or is_element_node(obj[0]):
+            elif isinstance(obj[0], tuple):
+                try:
+                    value = obj[0][1]
+                except IndexError:
+                    value = obj[0][0]
+                return True if isinstance(value, str) else bool(value)
+            elif is_element_node(obj[0]):
                 return True
             elif len(obj) == 1:
                 return bool(obj[0])
@@ -581,8 +644,10 @@ class XPathToken(Token):
         """
         if obj is None:
             return ''
+        elif is_schema_node(obj):
+            return str(self.schema_node_value(obj))
         elif is_element_node(obj):
-            return ''.join(elem_iter_strings(obj))
+            return ''.join(etree_iter_strings(obj))
         elif is_attribute_node(obj):
             return str(obj[1])
         elif is_text_node(obj):
@@ -595,8 +660,8 @@ class XPathToken(Token):
             return obj.text
         elif is_processing_instruction_node(obj):
             return obj.text
-        elif is_schema_node(obj):
-            return str(self.schema_node_value(obj))
+        elif isinstance(obj, bool):
+            return 'true' if obj else 'false'
         else:
             return str(obj)
 
@@ -615,17 +680,26 @@ class XPathToken(Token):
         of the node. Used for schema-based dynamic evaluation of XPath expressions.
         """
         try:
+            try:
+                return XSD_BUILTIN_TYPES[obj.type.local_name].value
+            except KeyError:
+                pass
+
             if obj.type.is_simple() or obj.type.has_simple_content():
                 # In case of schema element or attribute use a the sample value
                 # of the primitive type
                 primitive_type = self.parser.schema.get_primitive_type(obj.type)
-                return XSD_BUILTIN_TYPES[primitive_type.local_name].value
-            elif obj.type.local_name == 'anyType':
-                return XSD_BUILTIN_TYPES['anyType'].value
+                try:
+                    return XSD_BUILTIN_TYPES[primitive_type.local_name].value
+                except KeyError:
+                    return XSD_BUILTIN_TYPES['anyType'].value
             else:
                 return UntypedAtomic('')
+
         except AttributeError:
-            raise self.wrong_type("the argument %r is not a node of an XSD schema" % obj)
+            if self.parser.schema is None:
+                self.missing_schema()
+            self.wrong_type("the argument {!r} is not a node of an XSD schema".format(obj))
 
     ###
     # Error handling helpers
